#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Выбор декоратора асинхронных методов для превращения их в корутины

    По умлчанию используется декоратор из коробки types.coroutine.
    Он является ранней реализацией async|await и полностью совместим с этим
    механизмом

    Возможно глобальное переключение соответствующий декоратор из пакета
    Tornado

"""
import types

async_decorator = types.coroutine

try:
    # noinspection PyPackageRequirements
    from tornado.gen import coroutine as tornado_decorator
    # noinspection PyPackageRequirements
    from tornado.ioloop import IOLoop, PeriodicCallback
    # noinspection PyPackageRequirements
    from tornado.httpserver import HTTPServer
    # noinspection PyPackageRequirements
    from tornado.stack_context import run_with_stack_context, StackContext
except ImportError:
    tornado_decorator = None
    HTTPServer = None
    PeriodicCallback = None
    IOLoop = None
    run_with_stack_context = None
    StackContext = None


def switch_to_tornado():
    global async_decorator
    if tornado_decorator:
        async_decorator = tornado_decorator
    return tornado_decorator


def get_coroutine_decorator():
    return async_decorator


def is_tornado_on():
    return async_decorator == tornado_decorator



import re
from math import fabs

from tracuni.define.const import MASK_PARAMS


def mask_partial(row, symbol='*', from_idx=4, to_idx=-4):
    """
    Частичное экранирование строки
    :param row: строка для экранирования
    :param symbol: экранирующий символ
    :param from_idx: стартовая позиция маски
    :param to_idx: конечная позиция маски
    :return:
    """
    row = str(row)
    mask_len = round(len(row) - fabs(from_idx) - fabs(to_idx))
    masked_row = row[:from_idx] + mask_len * symbol + row[to_idx:]
    return masked_row


def mask_full(row, symbol='x'):
    """
    Полное экранирование строки
    :param row: строка
    :param symbol: экранирующий символ
    :return:
    """
    row = str(row)
    masked_row = symbol * len(row)
    return masked_row


def replace_fn(match_obj):
    """
        Пробуем сохранитьб валидный JSON:
            * Если мы число поменяли на строку, оборачиваем в кавычки
            * Если у нас были параметры запроса внутри строки, то возвращаем
            хвостовую кавычку
    :param match_obj:
    :return:
    """
    groups = match_obj.groups()

    variant = MASK_PARAMS.get(groups[0])

    mask_fn = None
    if variant == 'full':
        mask_fn = mask_full
    elif variant == 'part':
        mask_fn = mask_partial

    if mask_fn:
        masked = mask_fn(groups[2] or groups[5])
        double_quotes_tail = groups[3] or ''
        double_quotes = '"' if groups[5] else ''
        sep = groups[1] or groups[4]

        return (
            groups[0]
            +
            sep
            +
            double_quotes
            +
            masked
            +
            double_quotes
            +
            double_quotes_tail
        )


kwords = r'(' + r'|'.join(MASK_PARAMS.keys()) + r')'
p = re.compile(
    kwords + r'(?:(\s*?=\s*)([^&\n"]*)("?)|([\'"]\s*?:\s*)"?([^,\n"]*)"?)'
)


def do(data):
    data = p.sub(replace_fn, data)
    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Настраиваемые константные значения

"""

###############################################################################
# Значения, потенциально интересные пользователю


UNKNOWN_NAME = '<unknown>'
"""Такая строка будет отображаться там, где не удалось получить значение
"""

TRACER_SYSTEM_NAME = 'zipkin'
"""Такая строка должна быть в logging.tracer, чтобы данный трассер работал
"""

CUT_LIMIT = 4000
"""Максимальная длина строки содержимого

    Строки, длина которых в символах больше указанного здесь значения,
    будут обрезаться посередине
"""

CUT_LINE = '.......<CUT OVER {cut_limit} CHARS, FROM {ttl_len} TOTAL>.......'
"""Такая строка появится на месте вырезанного содержимого

    Можно использовать значения cut_limit, ttl_len. Или не использовать
"""

LOG_STRING_WRAP = 60
"""Аннотации являющиеся текстом не JSON переносить по строкам такой длины
"""

TAG_ESSENTIAL_FIELD_NAMES = (
    'form_request_id',
    ('payment_id', 'pid',),
    'link_id',
    'client_id',
    'refill_id',
    'account',
    'orderId',
    'orderStatus',
)
"""Существенные значения, помещаемые в заголовок

    Атрибуты, значения которых надо вытаскивать из тела запроса
    и помещать в заголовок, в метки для поиска
    если элемент - список, то вытаскивать значения под всеми этими именами,
    а размещать в метках под первым
"""

TAG_ESSENTIAL_BODY_NAMES = (
    'request_body',
    'response_body',
    'params',
    'pay_params',
    'params_extra',
)
"""Где искать существенные значения

    Если элементы данных представляют из себя словари или объекты,
    то существенные данные разыскиваются в их ключах или атрибутах
    соответственно под данными именами
"""

MASK_PARAMS = {
    'pan': 'part',
    'cvv': 'full',
    'cvc': 'full',
    'password': 'full',
    'secret': 'full',
    'secret_key': 'full',
}
"""Какие данные маскировать

    Перед записью в участок трассера, уже преобразованное в с троку 
    перехваченное значение прогоняется через регулярное выражение, которое 
    заменяет данные встречененые после присутствующих здесь слов на символ 
    'x', или на частичное маскирование звездочками в варианте 'part'
"""


###############################################################################
# Никому неинтересные внутренние значения


COLLECT_TRACE = True
"""Собирать ли список вызовов при прохождении перехватчика

    Из него берутся наименования методов, вызывающие перехватвыаемую точку, 
    чтобы в наименовании вместо ничего не говорящего "query" увидеть. 
    например, "get_checked_payment" или вместо "send" 
    получить "make_upay_request"
    
    Флаг был введён когда получение стека было реализовано накладным способом, 
    через встроенную функцию inspect, сейчас, возможно, он уже потерял свой 
    смысл
"""

PATH_IGNORE_METHODS = ('decorated', 'wrapper', '<module>')
"""Игнорировать методы для истории стека

    Для сбора истории стека вызовов игнорировать методы с данными именами
"""

PATH_IGNORE_PIECE = r'traceruni/'
"""Игнорировать пути для истории стека

    Для сбора истории стека вызовов игнорировать методы из модулей по данному
    пути
"""

CONTEXT_UID_NAME = 'span_context_uid'
"""Идентификатор участка в контексте

    Наименование ключа для идентификатора участка при сохранении в контексте
    асинхронных вызовов
"""

CONTEXT_SPAN_NAME = 'span_context'
"""Данные участка в контексте

    Наименование ключа для идентификатора участка при сохранении в контексте
    асинхронных вызовов
"""

SECRET_MASK_METHOD_NAME = 'pipe_secret'
"""Наименование маскирующего метода для работы конфигурации его запрета
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Перехват вызова интересующей точки
    Во время перехвата запускается процесс сбора, обработки и записи данных,
    согласно связаным с перехватчиком набором правил
"""
import inspect
import logging
from functools import wraps
from typing import (
    Optional,
    Callable,
)

from tracuni.adapter import TracerAdapter
from tracuni.schema.engine.engine import SchemaEngine
from tracuni.point.span_general import (
    SpanGeneral,
)
from tracuni.point.accessor import (
    PointAccessor,
    PointContext,
)
from tracuni.define.type import (
    Variant,
    RuleSet,
    Stage,
)
import tracuni.define.errors as err
from tracuni.misc.helper import PostgreSQLError
from tracuni.misc.select_coroutine import get_coroutine_decorator
async_decorator = get_coroutine_decorator()
try:
    import tornado
except ImportError:
    pass


def tracer_point(
        variant: Variant,
        remote_name: str = None,
        point_ruleset: RuleSet = None,
):
    """

    Parameters
    ----------
    variant
        Назначить вариант обработки точки
    remote_name
        Здесь можно передать пользовательское имя удаленной точки
        при этом можно использовать и стандартное, указав в строке место
        подстановки с именем standard: 'mine: {standard}'
    point_ruleset
        Здесь можно передать свой вариант настройки извлечения формат такой
        же как значение одного из ключей стандартной настройки она заменит
        стандартную настройку конкретного варианта для данной точки,
        если же последним элементом пользовательской настройки будет строка
        '...', то пользовательская объединится со стандартной с приоритетом
        пользовательской в любом случае далее удут применены менее
        специфичные и менее приоритетные настройки, относящиеся к данной точке

    Returns
    -------
        Декорированный метод, вызов которого будет трассироваться

    """
    # создание экземпляра обработчика
    schema_engine = SchemaEngine(variant, point_ruleset)
    # подписываемся, чтобы знать состояние трассера
    TracerAdapter.register_state_observer(schema_engine)

    def wrapper(point_coro):

        @wraps(point_coro)
        # @async_decorator
        @get_coroutine_decorator()
        def decorated(*args, **kwargs):
            f = inspect.currentframe()
            result = {}
            point = None
            point_fn = None
            this_span: Optional[SpanGeneral] = None
            debug_is_on = False

            try:
                # сразу запускаем обернутую точку, если трассер выключен,
                # не соединился или не создан
                if schema_engine.is_tracer_disabled():
                    result = yield from point_coro(*args, **kwargs)
                    return result

                debug_is_on = schema_engine.adapter.config.debug

                point = PointAccessor(
                    args,
                    kwargs,
                    PointContext(
                        point_ref=point_coro,
                        engine=schema_engine,
                        adapter=schema_engine.adapter,
                        client=schema_engine.adapter.client,
                        reuse={
                            "custom_remote_endpoint_name": remote_name,
                            "error": None,
                            "should_not_trace": False,
                        },
                        headers={},
                        span_name={},
                        span_tags={},
                        span_logs=[],
                    ),
                )
                point.stack_frames_to_context(inspect.currentframe())

                # создаем спан сконфигрированный по варианту и связанный с
                # ним и точкой
                # this_span = point.fab_span_general()
                # если не создан участок переходим к прямому вызову
                # if this_span:
                    # yield from this_span.get_stage_method(Stage.INIT)()
                    # # создаем участок, заполняем его данными доступными до
                    # # вызова получаем готовый обернутый метод для вызова точки
                    # pre_method = this_span.get_stage_method(Stage.PRE)
                    # point_fn = yield from pre_method()  # type: Callable

                yield from point.register_stage(Stage.INIT)
                if point.context.span:
                    yield from point.register_stage(Stage.PRE)

            except Exception as e:
                # если здесь вылетели, то рабочий участок создать не удалось
                # schema_engine.tracer_wrap.log_debug('Error in tracer: {
                # }'.format(e))
                # this_span = None
                if point:
                    point.register_own_error(e)
                if debug_is_on:
                    if not isinstance(
                        e,
                        (err.HTTPPathSkip, err.SpanNoParentException)
                    ):
                        logging.exception(e)
                    else:
                        logging.debug(e)

            finally:
                # в любом случае должны вызвать обернутую точку
                try:
                    result = yield from point.run()
                except (Exception, PostgreSQLError) as e:
                    point.register_error(e)
                    # фиксируем данные об исключении
                    if debug_is_on:
                        logging.exception(e)
                finally:
                    # так или иначе пробуем завершить формирование
                    # участка трассера
                    try:
                        yield from point.register_stage(Stage.POST)
                    except Exception as e:
                        if point:
                            point.register_own_error(e)
                        if debug_is_on:
                            logging.exception(e)

                if point.has_not_been_done():
                    # если вызов через аксессор не удался, вызываем напрямую
                    result = yield from point_coro(*args, **kwargs)

            return result

        return decorated

    return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Тесты модуля типов

"""

import sys
from enum import Enum

from typing import (
    Sequence,
    Optional,
    Union,
    Callable,
)

import pytest

from tracuni import trac_uni
from tracuni.trac_uni import TracerOptions, StatsdOptions

# pytest.skip('skipping {} tests'.format(__name__), allow_module_level=True)


###############################################################################
# Настройки трассера и сервиса сбора метрик


@pytest.fixture
def default_tracer_options(monkeypatch, unknown_name):
    """Настройки трассера со значениями по умолчанию
    """
    # if 'trac_uni.define.const' in sys.modules:
    #     del sys.modules['trac_uni.define.const']
    # mocker.patch('trac_uni.define.const.UNKNOWN_NAME', unknown_name)
    monkeypatch.setattr(
        trac_uni.define.type.TracerOptions.__new__,
        '__defaults__',
        (
            'zipkin', False, unknown_name, '', 1, 3, (), False,
            StatsdOptions(
                *StatsdOptions.__new__.__defaults__
            ),
        ),
    )
    return TracerOptions()


@pytest.fixture
def default_statsd_options():
    """Настройки сервиса метрик со значениями по умолчанию
    """
    from tracuni.trac_uni import StatsdOptions
    return StatsdOptions()


@pytest.fixture
def unknown_name():
    """Тестовое имя для импортируемого из модуля констант
    """
    return "some_ludicrous_service_name"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_tracer_default_options(
    monkeypatch,
    default_tracer_options,
    default_statsd_options,
    unknown_name,
    is_named_tuple_class,
):
    """Проверка настроек по умолчанию для трассера
    """
    # default_tracer_options = TracerOptions()
    assert is_named_tuple_class(type(default_tracer_options))
    assert default_tracer_options.name == 'zipkin'
    assert default_tracer_options.enable is False
    assert default_tracer_options.service_name == unknown_name
    assert default_tracer_options.url == ''
    assert default_tracer_options.sample_rate == 1
    assert default_tracer_options.send_interval == 3
    assert isinstance(default_tracer_options.skip_trace, Sequence)
    assert len(default_tracer_options.skip_trace) == 0
    assert default_tracer_options.debug is False
    assert hasattr(default_tracer_options, 'statsd')
    assert isinstance(
        default_tracer_options.statsd, type(default_statsd_options)
    )


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_statsd_default_options(default_statsd_options, is_named_tuple_class):
    """Проверка настроек по умолчанию для сервиса метрик
    """
    assert is_named_tuple_class(type(default_statsd_options))
    assert default_statsd_options.enable is False
    assert default_statsd_options.logging is False
    assert default_statsd_options.url == ''
    assert default_statsd_options.prefix == 'api'
    assert default_statsd_options.hb_interval == 2
    assert default_statsd_options.host is None
    assert default_statsd_options.port is None


###############################################################################
# Варианты участков


@pytest.fixture
def default_variant():
    """Создание экземпляра варианта точки по умолчанию
    """
    from tracuni.trac_uni import (
        Variant,
    )
    return Variant()


@pytest.fixture
def variant_types():
    """Импорт типов варианта точки
    """
    # if 'trac_uni.define.type' in sys.modules:
    #     del sys.modules['trac_uni.define.type']
    from tracuni.trac_uni import (
        Variant,
        SpanSide,
        APIKind,
    )
    return [Variant, SpanSide, APIKind]


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_auto_enum(variant_types):
    v_t, side_t, kind_t = variant_types
    assert side_t.IN < side_t.OUT
    with pytest.raises(TypeError) as exc_info:
        res = side_t.IN > 1
        del res
    assert exc_info


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_variant_type(variant_types, is_named_tuple_class):
    """Тип варианта содержит данные о направлении и виде интерфейса точки
    """
    variant_type, _, _ = variant_types
    assert is_named_tuple_class(variant_type)
    assert hasattr(variant_type, 'span_side')
    assert hasattr(variant_type, 'api_kind')


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_span_side_type(variant_types):
    """Существуют ожидаемые типы направления точки
    """
    _, span_side_type, _ = variant_types
    assert issubclass(span_side_type, Enum)
    assert hasattr(span_side_type, 'ALL')
    assert hasattr(span_side_type, 'SERVER')
    assert hasattr(span_side_type, 'CLIENT')
    assert hasattr(span_side_type, 'INCOMING')
    assert hasattr(span_side_type, 'OUTGOING')
    assert hasattr(span_side_type, 'IN')
    assert hasattr(span_side_type, 'OUT')


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_api_kind_type(variant_types):
    """Существуют ожидаемые типы интерфейса точки
    """
    _, _, api_kind_type = variant_types
    assert issubclass(api_kind_type, Enum)
    assert hasattr(api_kind_type, 'ALL')
    assert hasattr(api_kind_type, 'HTTP')
    assert hasattr(api_kind_type, 'AMQP')
    assert hasattr(api_kind_type, 'DB')


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_default_variant(default_variant):
    """По умолчанию создаётся вариант, относящийся ко всем разновидностям точек
    """
    assert str(default_variant.span_side) == 'SpanSide.ALL'
    assert str(default_variant.api_kind) == 'APIKind.ALL'


###############################################################################
# Информация по обернутой точке


@pytest.fixture
def point_context_type(mocker):
    """Импорт типа контекста точки
    """
    if 'trac_uni.type' in sys.modules:
        del sys.modules['trac_uni.type']
    mocker.patch('typing.TYPE_CHECKING', True)
    from tracuni.trac_uni import (
        PointContext,
    )
    return PointContext


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_point_context_type(point_context_type, is_named_tuple_class):
    """"Существует ожидаемый тип контекста точки
    """
    assert is_named_tuple_class(point_context_type)


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_point_context_instance_miss_required(point_context_type):
    """Должен валиться если не заданы все обязательные параметры
    """
    with pytest.raises(TypeError):
        point = point_context_type(
        )
        del point
    with pytest.raises(TypeError):
        point = point_context_type(
            POINT_ARGS={'a': 'a'}
        )
        del point


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_point_context_instance(point_context_type):
    """Можно создать экземпляр контекста точки
    """

    def test_func(*args, **kwargs):
        return kwargs['x'] + sum(args)

    point = point_context_type(
        point_args={'x': 10},
        point_ref=test_func,
        adapter=object(),
        engine=object(),
        span=object(),
    )
    assert point.point_ref(*[10, 22], **point.point_args) == 42


###############################################################################
# Структура конфигурации обработки наполнения участков


@pytest.fixture
def stage_type():
    """Импорт типа фаз извлечения содержимого
    """
    from tracuni.trac_uni import (
        Stage
    )
    return Stage


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_stage_type(stage_type):
    """Существуют ожидаемые виды фаз извлечения
    """
    assert issubclass(stage_type, Enum)
    assert hasattr(stage_type, 'INIT')
    assert hasattr(stage_type, 'PRE')
    assert hasattr(stage_type, 'POST')


@pytest.fixture
def origin_section_type():
    """Импорт типа секций источников извлекаемых данных
    """
    from tracuni.trac_uni import (
        OriginSection
    )
    return OriginSection


@pytest.fixture
def origin_type():
    """Импорт типа описания извлекаемых данных
    """
    from tracuni.trac_uni import (
        Origin
    )
    return Origin


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_origin_section_type(origin_section_type):
    """Существуют ожидаемые секции источников
    """
    assert issubclass(origin_section_type, Enum)
    assert hasattr(origin_section_type, 'POINT_ARGS')
    assert hasattr(origin_section_type, 'POINT_RESULT')
    assert hasattr(origin_section_type, 'POINT_REF')
    assert hasattr(origin_section_type, 'CLIENT')
    assert hasattr(origin_section_type, 'ADAPTER')
    assert hasattr(origin_section_type, 'ENGINE')
    assert hasattr(origin_section_type, 'CALL_STACK')
    assert hasattr(origin_section_type, 'REUSE')
    assert hasattr(origin_section_type, 'SPAN')


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_origin_type(origin_type, origin_section_type, is_named_tuple_class):
    """"Существует ожидаемый тип настройки источника извлекаемых данных
    """
    assert is_named_tuple_class(origin_type)
    assert hasattr(origin_type, 'section')
    assert origin_type.__annotations__['section'] == origin_section_type
    assert hasattr(origin_type, 'getter')
    assert origin_type.__annotations__['getter'] == Union[str, Callable]


@pytest.fixture
def destination_section_type():
    """Импорт типа секций назначения извлекаемых данных
    """
    from tracuni.trac_uni import (
        DestinationSection
    )
    return DestinationSection


@pytest.fixture
def destination_type():
    """Импорт типа наименования атрибута назначения
    """
    from tracuni.trac_uni import (
        Destination
    )
    return Destination


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_destination_section_type(destination_section_type):
    """Существуют ожидаемые секции назначения
    """
    assert issubclass(destination_section_type, Enum)
    assert hasattr(destination_section_type, 'TAGS')
    assert hasattr(destination_section_type, 'LOGS')
    assert hasattr(destination_section_type, 'SPAN_ID')
    assert hasattr(destination_section_type, 'ALL')


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_destination_type(destination_type,
                          destination_section_type,
                          is_named_tuple_class):
    """"Существует ожидаемый тип настройки назначения извлекаемых данных
    """
    assert is_named_tuple_class(destination_type)
    assert hasattr(destination_type, 'section')
    assert (
        destination_type.__annotations__['section'] == destination_section_type
    )
    assert hasattr(destination_type, 'name')
    assert destination_type.__annotations__['name'] == Optional[str]


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_destination_instance_miss_required(destination_type):
    """Должен валиться если не заданы все обязательные параметры
    """
    with pytest.raises(TypeError):
        destination = destination_type()
        del destination


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_destination_instance(destination_type, destination_section_type):
    """Можно создать экземпляр собранного содержимого
    """
    destination = destination_type(
        section=destination_section_type.TAGS,
    )
    assert (
        str(destination) == 'Destination(section=<DestinationSection.TAGS: '
                            '1>, name=None)'
    )


@pytest.fixture
def tee_types():
    """Импорт типа наименования атрибута назначения
    """
    from tracuni.trac_uni import (
        PipelineTee,
        TeeDescriptor,
    )
    return PipelineTee, TeeDescriptor


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_tee_type(tee_types, is_named_tuple_class):
    """Существует ожидаемый тип настройки ответвлённой записи
    """
    tee_type, tee_descriptor_type = tee_types
    assert is_named_tuple_class(tee_type)
    assert hasattr(tee_type, 'main_story')
    assert hasattr(tee_type, 'side_story')
    assert tee_type.__annotations__['side_story'] == tee_descriptor_type


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_tee_instance_miss_required(tee_types):
    """Должен валиться если не заданы все обязательные параметры
    """
    tee_type, _ = tee_types
    with pytest.raises(TypeError):
        tee = tee_type()
        del tee


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_tee_instance(
    tee_types,
    origin_type,
    origin_section_type,
    destination_type,
    destination_section_type,
):
    """Можно создать экземпляр настройки ответвлённой записи
    """
    tee_type, _ = tee_types
    tee = tee_type(
        main_story='11',
        side_story={
            origin_type(origin_section_type.CLIENT, 'headers'): {
                'header': 'content'
            }
        }
    )
    del tee


@pytest.fixture
def stage_maps():
    """Импорт типа секций назначения извлекаемых данных
    """
    from tracuni.trac_uni import (
        stage_to_sections,
        method_to_stage,
        Stage,
    )
    expected_keys = (Stage.INIT, Stage.POST)
    return method_to_stage, stage_to_sections, expected_keys


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_method_stage_map(stage_maps):
    method_stage_map, stage_sections_map, expected_keys = stage_maps
    assert type(method_stage_map) == dict
    assert type(stage_sections_map) == dict
    assert not set(stage_sections_map.keys()) ^ set(expected_keys)


###############################################################################
# Настройка обработки наполнения участков


@pytest.fixture
def extractables_types():
    """Импорт типа обработки
    """
    from tracuni.trac_uni import (
        Rule,
        Origins,
        Pipeline,
        Schema,
    )
    return Rule, Origins, Pipeline, Schema


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_extractables_type(extractables_types,
                           stage_type,
                           destination_type,
                           is_named_tuple_class):
    """Существует ожидаемый тип настройки настройки извлечения данных участка
    """
    extract_type, origins_type, extract_proc, _ = extractables_types
    assert is_named_tuple_class(extract_type)
    assert hasattr(extract_type, 'destination')
    assert extract_type.__annotations__['destination'] == destination_type
    assert hasattr(extract_type, 'origins')
    assert extract_type.__annotations__['origins'] == origins_type
    assert hasattr(extract_type, 'pipeline')
    assert extract_type.__annotations__['pipeline'] == extract_proc
    assert hasattr(extract_type, 'description')
    assert extract_type.__annotations__['description'] == Optional[str]
    assert hasattr(extract_type, 'stage')
    assert extract_type.__annotations__['stage'] == Optional[stage_type]


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test_extractables(extractables_types,
                      variant_types,
                      stage_type,
                      destination_type,
                      destination_section_type,
                      origin_type,
                      origin_section_type,
                      ):
    """Создаются разные варианты настройки извлечения данных участка
    """
    extract_type, _, _, setup_type = extractables_types
    variant_type, span_side_type, api_kind_type = variant_types

    setup_empty = {

    }  # type: setup_type
    assert str(setup_empty) == '{}'
    del setup_empty

    setup_some = {  # type: setup_type
        variant_type(): (
            extract_type(
                description='test',
                destination=destination_type(
                    section=destination_section_type.LOGS,
                    name='some'
                ),
                origins=(
                    origin_type(
                        section=origin_section_type.SPAN,
                        getter='id'
                    )
                ),
                pipeline=(),
            ),
        )
    }
    assert isinstance(setup_some, dict)
    del setup_some

    setup_all = {  # type: setup_type
        variant_type(): (
            extract_type(
                destination=destination_type(
                    section=destination_section_type.LOGS,
                    name='some'
                ),
                origins=(
                    origin_type(
                        section=origin_section_type.POINT_ARGS,
                        getter=lambda v: str(v)
                    )
                ),
                pipeline=(),
            ),
            extract_type(
                description="some",
                destination=destination_type(
                    section=destination_section_type.TAGS,
                    name='some'
                ),
                origins=(
                    origin_type(
                        section=origin_section_type.SPAN,
                        getter='id'
                    )
                ),
                pipeline=(),
            ),
            extract_type(
                destination=destination_type(
                    section=destination_section_type.SPAN_ID,
                    name='some'
                ),
                origins=(
                    origin_type(
                        section=origin_section_type.POINT_RESULT,
                        getter='id'
                    ),
                    origin_type(
                        section=origin_section_type.POINT_REF,
                        getter='id'
                    ),
                    origin_type(
                        section=origin_section_type.CLIENT,
                        getter='id'
                    ),
                ),
                pipeline=(),
                stage=stage_type.POST,
            ),
            extract_type(
                destination=destination_type(
                    section=destination_section_type.TAGS,
                    name='some'
                ),
                origins=(
                    origin_type(
                        section=origin_section_type.REUSE,
                        getter='id'
                    ),
                    origin_type(
                        section=origin_section_type.ADAPTER,
                        getter='id'
                    ),
                    origin_type(
                        section=origin_section_type.ENGINE,
                        getter='id'
                    ),
                ),
                pipeline=(
                    lambda *args: ','.join(str(el) for el in args)
                ),
            ),
            extract_type(
                destination=destination_type(
                    section=destination_section_type.LOGS,
                    name='some'
                ),
                origins=(
                    origin_type(
                        section=origin_section_type.CALL_STACK,
                        getter='id'
                    )
                ),
                pipeline=(),
            ),
            extract_type(
                destination=destination_type(
                    section=destination_section_type.TAGS,
                    name='some'
                ),
                origins=(
                    origin_type(
                        section=origin_section_type.SPAN,
                        getter='id'
                    )
                ),
                pipeline=(),
            ),
        )
    }
    assert isinstance(setup_all, dict)
    del setup_all



import pytest


pytest_plugins = [
    "tests.fixtures.fixture_engine_schemas",
    "tests.fixtures.fixture_engine_structures",
    "tests.fixtures.fixture_engine",
]


@pytest.fixture
def is_named_tuple_class():
    def _(t):
        b = t.__bases__
        if len(b) != 1 or b[0] != tuple:
            bb = b[0].__bases__
            if len(bb) != 1 or bb[0] != tuple:
                return False
        f = getattr(t, '_fields', None)
        if not isinstance(f, tuple):
            return False
        return all(type(n) == str for n in f)

    return _



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Обработчик схем

    Управление схемами извлечения данных для различных вариантов точек

    Используется три уровня схем (в порядке убывания приоритета):
        * пользовательская на конкретную точку
            передаётся как инициализирующий параметр декоратора
        * пользовательская карта схем
            передаётся от пользователя как инициализирующий параметр
            адаптера библиотеки. Передаётся в процессор в момент его
            создания при инициализации декоратора
        * стандартная карта схем
            импортируется из schema.standard
"""

from collections import namedtuple
import inspect

from typing import (
    TYPE_CHECKING,
    Any,
    Tuple,
    Generator,
    MutableSequence,
    Callable)

from tracuni.define.type import (
    Rule,
    DestinationSection,
    PipelineTee,
    PointContext,
)
from tracuni.define import errors as err
from tracuni.misc.select_coroutine import async_decorator

from .cache import SchemaEngineCache
from .feed import SchemaEngineFeed

if TYPE_CHECKING:
    from typing import List  # noqa
    from typing import Union  # noqa
    from typing import Optional  # noqa
    from tracuni.define.type import Destination  # noqa
    from tracuni.define.type import MethodNameToStage  # noqa
    from tracuni.define.type import (  # noqa
        RuleSet,
        Extractors,
    )
    from tracuni.define.type import Stage
    from tracuni.adapter import AdapterType  # noqa


class SchemaEngine(metaclass=SchemaEngineCache):
    """Обработчик схемы. Связь варианта со схемой извлечения данных

        * Обращается к экземпляру поставщика схем для получения списка
        применяемых правил, разбитых на фазы и дополнительных данных
        * Для заданного экземпляра объекта доступа к точке извлекает и
        возвращает данные согласно правилам и состоянию объекта доступа.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.schema_feed = None  # type: SchemaEngineFeed
        self.adapter = None  # type: AdapterType
        self.extractors = None  # type: Extractors

        self.method_to_stage = None  # type: Optional[MethodNameToStage]
        self.skip_secret_destinations = (
            None  # type: Optional[MutableSequence[Destination]]
        )
        self.secret_mask_method_name = None  # type: Optional[str]

        self._args = args
        self._kwargs = kwargs

    def observe_state_change(self, adapter: 'AdapterType'):
        """
            Обработчик схемы создаётся до того, как отработает фабрика адаптера,
            соответственно вторая часть инициализации должна проходить после
            того, как адаптер будетсоздан

        Parameters
        ----------
        adapter
            только что созданный экземпляр адаптера
        """
        self.adapter = adapter
        self._kwargs['custom_schema'] = adapter.custom_schema
        self.schema_feed = schema_feed = SchemaEngineFeed(
            *self._args,
            **self._kwargs,
        )
        self.extractors = schema_feed.parse()  # type: Extractors

        self.skip_secret_destinations = (
            schema_feed.get_skip_secret()
        )  # type: MutableSequence[Destination]

        self.secret_mask_method_name = (
            schema_feed.get_secret_mask_method_name()
        )  # type: str

    def is_tracer_disabled(self):
        return self.adapter is None

    @async_decorator
    def extract(self,
                point_context: PointContext,
                stage: 'Stage',
                ) -> Generator[Any, None, PointContext]:
        """Метод запускающий процесс получения и изменения исходных данных и
        данных участка для указанной точки и этапа

        Parameters
        ----------
        point_context
            Ссылки на данные точки и окружения с ключами, совпадающими с
            наименованиями секций источников и назначения.
        stage
            Фаза, к которой привязан обратившийся метод

        Returns
        -------
            Изменённый контекст точки
        """
        if stage is None:
            return point_context
        extractors = (
            self.extractors.get(stage) or []
        )  # type: Union[RuleSet, List]

        for extractor in extractors:
            extracted = yield from self._extract_values(
                extractor,
                point_context
            )
            if extracted is None or not len(extracted):
                continue
            point_context = self._pipe_values(
                extractor,
                point_context,
                extracted,
            )

        return point_context

###############################################################################
# Area privata

    @staticmethod
    @async_decorator
    def _extract_values(
        extractor: Rule,
        point_context: PointContext
    ) -> Generator[Any, None, Tuple[Any, ...]]:
        """

        Parameters
        ----------
        extractor
            Применяемое в данный момент правило извлечения.
        point_context
            Структура, хранящая ссылки на контекст источник данных и уже
            собранные данные. Нужна для получения контекста источника.
        Returns
        -------
            Собранные исходные значения
        """
        val = []
        # правило может содержать несколько источников
        for origin in extractor.origins:
            extracted = None

            # получаем ссылку на источник или пропускаем источник
            section_ref = getattr(
                point_context, origin.section.name.lower(), None
            )
            if section_ref is None:
                continue

            # метод извлечения, либо функция, которой передается ссылка на
            # источник, а она возвращает значение по своему разумению
            getter = origin.getter
            if callable(getter):
                is_yields = inspect.isgeneratorfunction(getter)
                if is_yields:
                    extracted = yield from getter(section_ref)
                else:
                    extracted = getter(section_ref)
            # либо метод извлечения может быть указан строкой, тогда это или
            # элемент словаря, или атрибут объекта
            elif isinstance(getter, str):
                # отдельный случай прямой слэш - возращается исходный объект
                if getter == '/':
                    extracted = section_ref
                elif isinstance(section_ref, dict):
                    extracted = section_ref.get(getter)
                else:
                    extracted = getattr(section_ref, getter, None)
            elif getter is None:
                extracted = section_ref

            # if extracted is None:
            #     continue
            val.append(extracted)

        # данные, полученные из всех указанных в правиле источников,
        # собираются в кортеж
        return tuple(val)

    def _pipe_values(
        self,
        extractor: Rule,
        point_context: PointContext,
        val: Any,
    ) -> PointContext:
        """Конвейер обрботки извлекаемых значений

            Вызывает цепочку методов передавая от одного к другому данные в
            текущем состоянии, обрабатывает дополнительную команду ответвления
            данных и список подавленных обработчиков.

        Parameters
        ----------
        extractor
            Правило извелечения
        val
            Данные в текущем состоянии
        point_context
            Структура, хранящая ссылки на контекст источник данных и уже
            собранные данные.
        Returns
        -------
            Изменённый контекст точки
        """

        for process in extractor.pipeline:
            if (
                callable(process)
                and
                not self._skip_secret(process, extractor.destination)
            ):
                new_val = process(val)

                if isinstance(new_val, PipelineTee):
                    for section, new_values in new_val.side_story.items():
                        s_name = section.name.lower()
                        old_values = getattr(point_context, s_name, None)
                        if isinstance(old_values, dict):
                            old_values.update(new_values)
                        elif isinstance(old_values, list):
                            old_values.append(new_values)
                        else:
                            # noinspection PyProtectedMember
                            point_context = PointContext(**{
                                **point_context._asdict(),
                                **new_values,
                            })
                    new_val = new_val.main_story
                val = new_val

        point_context = self._update_point_context(
            extractor,
            point_context,
            val
        )
        return point_context

    @staticmethod
    def _update_point_context(
        extractor: Rule,
        point_context: PointContext,
        val: Any,
    ) -> PointContext:
        """

        Parameters
        ----------
        extractor
            Правило из которого берутся данныи по назначению записи
        point_context
            Модифицируемый контекст точки, вернется к держателю точки и
            будет обработан на его стороне
        val
            Значение которое надо отправить в контекст

        Returns
        -------
            Изменённый контекст точки
        """

        # noinspection PyProtectedMember
        dest = point_context._asdict().get(
            extractor.destination.section.name.lower()
        )
        if dest is None:
            return point_context
        elif isinstance(dest, MutableSequence):
            dest.append(val)
        elif isinstance(dest, dict):
            dest[extractor.destination.name] = val
        elif "PointArguments" in str(type(dest)):
            # noinspection PyProtectedMember
            old_vals = dest._asdict()
            new_args = namedtuple("PointArguments", old_vals)(**{
                **old_vals,
                extractor.destination.name: {
                    **(old_vals.get(
                        extractor.destination.name,
                        {}
                    ) or {}),
                    **val,
                }
            })
            # noinspection PyProtectedMember
            point_context = PointContext(**{
                **point_context._asdict(),
                "point_args": new_args,
            })
        else:
            try:
                if extractor.destination.name:
                    setattr(dest, extractor.destination.name, val)
            except AttributeError:
                raise err.ProcessorExtractCantSetDestinationAttr

        return point_context

    def _skip_secret(self, fn: Callable, destination: "Destination") -> bool:
        """Проверка условия пропуска обработчика маскирования по команде

        Parameters
        ----------
        fn
            Обработчик
        destination
            Описание назначения для проверки применения команды

        Returns
        -------
            Флаг надо ли запускать данный обработчик
        """
        fn_name = getattr(fn, '__name__', '')
        if self.secret_mask_method_name != fn_name:
            return False

        return (
            (
                Destination(DestinationSection.ALL, "")
                in
                self.skip_secret_destinations
            )
            or
            destination in self.skip_secret_destinations
        )



import asyncio
import hashlib
import hmac
import json
import os
import ssl
import sys
from functools import partial
from uuid import uuid4
from json import JSONDecodeError
from urllib.parse import urlencode

from app.core.handler import BaseHandler
from app.core.helper import json_encode
from app.core.http_client import HttpClient


class AmqpHandler(BaseHandler):
    routes = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.http_cln: HttpClient = self.app.http_cln
        self.routes = [
            {
                'route_name': self.config['rabbit']['queue_name'],
                'queue_name': self.config['rabbit']['queue_name'],
                'routing_key': self.config['rabbit']['routing_key'],
                'handler': self.receive
            }
        ]

    async def start(self):
        # перезапуск не завершенных задач
        pass

    async def receive(self, channel, body, envelope, properties):
        self.app.log_info('[i] Request: %s' % body)

        try:
            req = json.loads(body.decode())

            self.app.log_info("[i] Received %r" % req)

            if req.get('name') == 'cb':
                self.app.log_info("Result: {}".format(
                    json_encode(req)
                ))
            else:

                cb_type = None
                cb_address = None

                waiting_for = []
                for name, task in req.get('tasks', {}).items():
                    kind = task.get('api_kind')
                    sub_tasks = task.get('tasks', {})
                    if name == "cb":
                        cb_type = kind
                        cb_address = task.get('address', '')
                    else:
                        if task.get('api_kind') == 'http':
                            url = task.get('address')
                            if url:
                                http_res = self.app.http_cln.request(
                                    "POST",
                                    url,
                                    headers={},
                                    body=json.dumps({
                                        "name": name,
                                        "tasks": sub_tasks
                                    }).encode(),
                                )
                                waiting_for.append(http_res)
                        if task.get('api_kind') == 'amqp':
                            amqp_res = self.task(name, task)
                            waiting_for.append(amqp_res)
                        if task.get('api_kind') == 'db':
                            db_res = getattr(self.app.db_tasks,
                                             task.get('address'))(name, task)
                            waiting_for.append(db_res)

                if waiting_for:
                    response = await asyncio.gather(*waiting_for)
                else:
                    response = []
                response = {
                    "name": req.get("name", "unknown"),
                    "ok": True,
                    "tasks": dict(
                        (
                            response_item.get("name", uuid4().hex),
                            dict(
                                (k, v)
                                for k, v in response_item.items()
                                if k != "name"
                            )
                        )
                        for response_item in response
                    ),
                }
                if cb_type == "http":
                    self.app.http_cln.request(
                        "POST",
                        cb_address,
                        headers={},
                        body=json.dumps(response).encode(),
                    )
                if cb_type == "amqp":
                    await self.task("cb", {
                        "address": cb_address,
                        "tasks": response,
                    })

        except JSONDecodeError:
            self.app.log_err("[x] Error decoding task %r" % body.decode())
        await channel.basic_client_ack(delivery_tag=envelope.delivery_tag)

    async def async_task_out(self, tid, data):
        url = 'wrong'

        res_http = await self.app.http_cln.request(
            data['method'],
            url,
            body={},
            headers={},
        )

        # noinspection PyPep8
        try:
            res_body = json.loads(res_http['response'])
            check_code_out = data['params_extra'].get('check_code_out', True)
            if check_code_out:
                res_code = int(res_body['code'])
            else:
                res_code = 0
        except:
            res_body = None
            res_code = 1

        self.app.log_info('[i] Task %i Res body %s' % (tid, json.dumps(res_body)))

        # проверка сответсвия http кода списку
        if res_http:
            http_code = res_http.get('http_code')
        else:
            http_code = 111




from functools import partial

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
)
from tracuni.schema.builtin.pipe import (
    ext_http_in_aiohttp_request,
    pipe_head,
    pipe_check_skip,
    pipe_sep_string,
    pipe_catch_essential,
    pipe_mask_secret,
    pipe_dump,
    pipe_cut,
)


ruleset = (
    Rule(
        description="Получить данные запроса и сохранить"
                    "для переиспользования",
        destination=Destination(DestinationSection.REUSE, 'request'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.POINT_ARGS, ext_http_in_aiohttp_request),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Поставить флаг пропуска трассировки, если HTTP путь"
                    "в списке игнорируемых",
        destination=Destination(DestinationSection.REUSE, 'should_not_trace'),
        pipeline=(
            pipe_check_skip,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['url_path'],
            ),
            Origin(OriginSection.ADAPTER, None),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Разместить заголовки контекста трассера там, "
                    "где фабрика участков их ищет",
        destination=Destination(DestinationSection.REUSE, 'tracer_headers'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['zipkin_headers'],
            ),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Задать имя участку",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'name'
        ),
        pipeline=(
            partial(pipe_sep_string, **{'sep': ' '}),
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['method'],
            ),
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['url_path'],
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Задать имя партнёрской точки для участка",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'remote_endpoint',
          ),
        pipeline=(
            partial(pipe_sep_string, **{'sep': '://'}),
        ),
        origins=(
            Origin(
                OriginSection.ENGINE,
                lambda engine: engine.schema_feed.variant.api_kind.name.lower(),
            ),
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['remote_ip'],
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести тело запроса в аннотации участка ",
        destination=Destination(DestinationSection.SPAN_LOGS, 'request_body'),
        pipeline=(
            pipe_head,
            # pipe_catch_essential,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret,
        ),
        origins=(
            Origin(OriginSection.REUSE, lambda data: data['request']['body']),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести параметры запроса в метки участка ",
        destination=Destination(DestinationSection.SPAN_TAGS, 'req.query'),
        pipeline=(
            pipe_head,
            # pipe_catch_essential,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret,
        ),
        origins=(
            Origin(OriginSection.REUSE, lambda data: data['request']['query']),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Сохранить информацию об ответе",
        destination=Destination(DestinationSection.REUSE, 'response'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.POINT_RESULT, None),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести тело ответа на запрос",
        destination=Destination(DestinationSection.SPAN_LOGS, 'response_body'),
        pipeline=(
            pipe_head,
            lambda response: getattr(response, 'body', b'').decode(),
            pipe_catch_essential,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Запомнить ошибку для вывода",
        destination=Destination(DestinationSection.REUSE, 'error'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.SPAN, 'error'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в тэги адрес запроса",
        destination=Destination(DestinationSection.SPAN_TAGS, 'url'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, lambda data: data['request']['url']),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в тэги размер запроса",
        destination=Destination(DestinationSection.SPAN_TAGS, 'http.req.size'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['length'],
            ),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в тэги заголовки запроса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'http.req.headers',
        ),
        pipeline=(
            pipe_head,
            pipe_dump,
        ),
        origins=(
            Origin(
                OriginSection.REUSE,
                lambda data: data['request']['headers'],
            ),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в тэги код статуса ответа",
        destination=Destination(DestinationSection.SPAN_TAGS,
                                'http.rsp.status'),
        pipeline=(
            pipe_head,
            lambda response: getattr(response, 'status_code', {}),
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в тэги размер ответа",
        destination=Destination(DestinationSection.SPAN_TAGS, 'http.rsp.size'),
        pipeline=(
            pipe_head,
            lambda response: len(getattr(response, 'body', b'').decode()),
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести в тэги заголовки ответа",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'http.rsp.headers',
        ),
        pipeline=(
            pipe_head,
            lambda response: getattr(response, 'headers', {}),
            lambda headers: dict(headers) if headers else None,
            pipe_dump,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'response'),
        ),
        stage=Stage.POST,
    ),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Общие методы создания и наполнение участков
"""
from aiozipkin.span import Span
import aiozipkin as az

from typing import (
    TYPE_CHECKING,
    Any,
)

from tracuni.misc.select_coroutine import async_decorator
from tracuni.define.type import (
    SpanSide,
    UNKNOWN_NAME,
    Stage,
)
import tracuni.define.errors as err
from tracuni.point.span_stages import ProvideSpanForStage

if TYPE_CHECKING:
    from tracuni.point.accessor import PointAccessor


class SpanGeneral:
    """Общая функциональность передачи данных на трассер

        * Реализует логику создания разных вариантов участков:
            * Входящий / Исходящий
            * С родительским контекстом / Без него
        * Записывает подготовленные данные в участок
            * В наименование
            * В заголовок
            * В аннотации

    """
    def __init__(self, side: SpanSide, point: 'PointAccessor'):
        self.span_instance = None
        self.side = side
        self.point = point
        self.tracer: az.Tracer = point.context.adapter.tracer
        self.error = None
        self.has_been_started = False

###############################################################################
# Создание участков для каждого из двух направлений
# Задействуются декоратором ProvideSpanForStage

    def fab_in_span(self):
        # если в перехватываемых данных есть загловки трасера,
        # создаем дочерний входящий участок,
        # иначе создаем корневой входящий участок
        headers = self.point.context.reuse.get("tracer_headers", {})
        context = az.make_context(headers)
        span_instance: Span = (
            self.tracer.new_trace()
            if context is None else
            self.tracer.new_child(context)
        )
        span_instance.kind(az.SERVER)
        # сохраняем идентификаторы входящего участка
        # все исходящие участки в данном дереве асинхронных вызовов
        # будут использовать его в качестве родительского контекста
        self.point.context.adapter.keep_context.set(
            'span',
            span_instance.context.make_headers(),
        )

        return span_instance

    def fab_out_span(self):
        # если в контексте нет входящего участка,
        # то исходящая точка вызывается вне сценария трассровки,
        # возвращаемся в перехватчик и делаем прямой вызов без трассировки
        in_span_context = self.point.context.adapter.keep_context.get('span')
        if in_span_context is None:
            raise err.SpanNoParentException
        # исходящий участок всегда создаётся как дочерний к входящему
        in_span_context = az.make_context(in_span_context)
        span_instance = self.tracer.new_child(in_span_context)
        span_instance.kind(az.CLIENT)

        return span_instance

    def enrich_headers(self, headers: dict, prefix_key=None):
        tracer_headers = self.span_instance.context.make_headers()
        if prefix_key:
            tracer_headers = {
                prefix_key: tracer_headers
            }
        headers.update(tracer_headers)
        return headers

###############################################################################
# Методы прохождения по этапу

    @ProvideSpanForStage(Stage.INIT)
    @async_decorator
    def run_init_stage(self, span: Span):
        self.has_been_started = True
        span.start()
        self._load()

    @ProvideSpanForStage(Stage.PRE)
    @async_decorator
    def run_pre_stage(self, _: Span):
        if not self.has_been_started:
            raise err.SpanIsNotStartedException
        self._load()

        self.point.apply_point()

    @ProvideSpanForStage(Stage.POST)
    @async_decorator
    def run_post_stage(self, span: Span):
        if not self.has_been_started:
            raise err.SpanIsNotStartedException
        self._load()
        span.finish(exception=self.error)

    @async_decorator
    def run_stage(self, stage: Stage):
        stage_method_name = ProvideSpanForStage.stage_to_method.get(stage)
        stage_method = getattr(self, stage_method_name or "", None)
        if stage_method:
            yield from stage_method()

    def register_error(self, e: Exception):
        self.error = e

    def emergency_exit(self, e):
        if self.span_instance:
            self.span_instance.finish(exception=e)

###############################################################################
# Area privata

    def _load(self):
        span = self.span_instance
        span_name = self.point.context.span_name
        span_tags = self.point.context.span_tags
        span_logs = self.point.context.span_logs

        span.name(span_name.get('name', UNKNOWN_NAME))
        custom_remote = self.point.context.reuse['custom_remote_endpoint_name']
        standard_remote = span_name.get('remote_endpoint', UNKNOWN_NAME)
        remote = (
            custom_remote.format(**{'standard': standard_remote})
            if custom_remote else
            standard_remote
        )
        span.remote_endpoint(remote)

        if span_logs is None:
            span_logs = []
        if span_tags is None:
            span_tags = {}

        for log_item in span_logs:
            if log_item and log_item != '{}':
                self.span_instance.annotate(log_item)
        self.point.update_context({"span_logs": []})
        for tag_name, tag_value in span_tags.items():
            if tag_value:
                self.span_instance.tag(tag_name, tag_value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Обработка участков на каждом этапе
"""
from functools import wraps

from tracuni.misc.select_coroutine import async_decorator
from tracuni.define.type import (
    SpanSide,
    Stage,
)
import tracuni.define.errors as err

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from aiozipkin.span import Span
    from typing import Dict
    from tracuni.define.type import MethodNameToStage


class ProvideSpanForStage:
    """Декоратор для методов этапа
            * Гарантирует наличие экземпляра участка
            * Получает наполнение для текущей стадии
            * Сохранят результат вызова точки

        Parameters
        ----------
        stage
            Привязывает декорированный метод к конкретному этапу. Если таких
            несколько, то привязывается только первый обработанный

        Returns
        -------
            Метод обработки определённого этапа
    """
    # Отношение наименований методов аксессора точки к запускаемым фазам
    # извлечения, отрабатывающим до выполнения этих методов.
    method_to_stage = {
        # 'run_init_stage': Stage.INIT,
        # 'run_pre_stage': Stage.PRE,
        # 'run_post_stage': Stage.POST,
    }  # type: MethodNameToStage

    stage_to_method = {}  # type: Dict[Stage, str]

    def __init__(self, stage: Stage):
        self.stage = stage

    def __call__(self, fn):
        # Сохраняем прямое и обраное отображение этапов на методы обработки
        self.fn_name = fn_name = fn.__name__
        if fn_name and self.stage not in self.method_to_stage:
            self.method_to_stage[fn_name] = self.stage
            self.stage_to_method[self.stage] = fn_name

        @wraps(fn)
        @async_decorator
        def wrapper(span_general_inst):
            self.span_general_inst = span_general_inst
            self.point = span_general_inst.point

            stage_found = self._detect_stage()
            self._stage_operations(stage_found)
            yield from self._run_stage_pipe(stage_found)
            span_instance = self._create_span_inst_if_none()

            fn(
                span_general_inst,
                span_instance,
            )

            return

        return wrapper

###############################################################################
# Area privata

    def _detect_stage(self):
        # Определяем текущий этап
        stage_found = self.method_to_stage.get(self.fn_name)
        if not stage_found:
            raise err.SpanNoStageForMethod

        return stage_found

    def _stage_operations(self, stage_found):
        # Для этапа после вызова точки сохраняем результат её работы
        if stage_found == Stage.POST:
            result = None
            if (
                not result
                and
                hasattr(
                    self.point.context.point_args,
                    'self',
                )
                and
                hasattr(
                    self.point.context.point_args.self,
                    'response_body',
                )
            ):
                result = self.point.context.point_args.self.response_body
            if self.point.context.point_result is None:
                self.point.update_context({"point_result": result})

    @async_decorator
    def _run_stage_pipe(self, stage_found):
        # Вызываем извлечение и обработку данных этапа
        pipe_fn = self.point.context.engine.extract
        point_context = yield from pipe_fn(self.point.context, stage_found)
        self.point.context = point_context

        # отмечаем, если произошла ошибка
        error = point_context.reuse.get('error')
        if error:
            self.span_general_inst.error = str(error)

    def _skip_http(self):
        # обрабатываем выставленный обработчиками флаг пропуска
        if self.point.context.reuse.get('should_not_trace'):
            raise err.HTTPPathSkip

    def _create_span_inst_if_none(self):
        # создаём участок, если он ещё не создан
        span_instance: Span = getattr(
            self.span_general_inst,
            'span_instance',
            None
        )
        if span_instance is None:
            if self.span_general_inst.side == SpanSide.IN:
                span_instance = self.span_general_inst.fab_in_span()
            elif self.span_general_inst.side == SpanSide.OUT:
                span_instance = self.span_general_inst.fab_out_span()

            if span_instance is None:
                raise err.NoSpanException
            self.span_general_inst.span_instance = span_instance

        if not self.point.context.span:
            self.point.update_context({"span": self.span_general_inst})

        return span_instance



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Привязка наборов правил к вариантам

    variant => ruleset

    ruleset = (
        Rule(
            description='',
            destination=Destination(
                section=DestinationSection.SPAN_TAGS,
                name='some'
            ),
            stage=Stage.PRE,
            origins=(
                Origin(section=OriginSection.REUSE, getter='some'),
            ),
            pipeline=(lambda x: x,),
        ),
    )  # type: RuleSet
"""
from typing import TYPE_CHECKING

from tracuni.misc.select_coroutine import is_tornado_on
from tracuni.define.type import (
    Variant,
    SpanSide,
    APIKind,
)


if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from ..define.type import RuleSet  # noqa
    from ..define.type import Schema  # noqa

####################
# Подключение стандартных настроек экстракторов

# для варианта будет использован только один из наборов для определённого
# ключа назначения в зависимости от разновидности ключа экстракторы
# применяются в следующем приоритете сверху вниз (если применяется верхний,
# то нижний уже не применяется)
#  - элементы с конкретным вариантом применяются
# только для этого варианта
#  - элементы привязанные только к направлению применяются ко всем вариантам
#  с этим направлением
#  - элементы привязанные по виду API применяются ко всем вариантам с этим
#  видом
#  - непривязанные к вариантам элементы применяются ко всем вариантам
amqp_in_ruleset = tuple()


def get_standard_schema():
    if is_tornado_on():
        from tracuni.schema.builtin import (
            tornado_http_in_ruleset as http_in_ruleset,
            tornado_http_out_ruleset as http_out_ruleset,
            tornado_db_out_ruleset as db_out_ruleset,
            tornado_amqp_out_ruleset as amqp_out_ruleset,
        )
    else:
        from tracuni.schema.builtin import (
            http_in_ruleset,
            http_out_ruleset,
            db_out_ruleset,
            amqp_out_ruleset,
        )
    standard_schema = {
        Variant(SpanSide.IN, APIKind.HTTP): http_in_ruleset,
        Variant(SpanSide.OUT, APIKind.HTTP): http_out_ruleset,
        Variant(SpanSide.IN, APIKind.AMQP): amqp_in_ruleset,
        Variant(SpanSide.OUT, APIKind.AMQP): amqp_out_ruleset,
        Variant(SpanSide.OUT, APIKind.DB): db_out_ruleset,
    }  # type: Schema
    return standard_schema



import re

import aiozipkin.transport as azt
import aiozipkin.constants as azc
import logging

from .adapter import StatsD
from tracuni.misc.select_coroutine import async_decorator


STATS_CLEAN_NAME_RE = re.compile('[^0-9a-zA-Z_.-]')
STATS_CLEAN_TAG_RE = re.compile('[^0-9a-zA-Z_=.-]')


class TracerTransport(azt.Transport):
    # noinspection PyProtectedMember
    def __init__(
            self,
            logging_conf,
            statsd_conf,
            loop,
            use_tornado=False,
    ):
        tracer_url = logging_conf.url
        send_interval = logging_conf.send_interval
        super(TracerTransport, self).__init__(
            tracer_url,
            send_interval=send_interval,
            loop=loop
        )
        self.loop = loop
        self.__tracer = logging_conf.name
        self.use_tornado = use_tornado

        statsd_conf = StatsD.read_statsd_configuration(statsd_conf)

        self.stats = None
        if statsd_conf.enable:
            self.stats = StatsD(statsd_conf._asdict())
            self.stats.connect()
        self.statsd_is_configured = statsd_conf.enable and self.stats

    @async_decorator
    def _send(self):
        data = self._queue[:]

        try:
            if self.statsd_is_configured:
                yield self._send_to_statsd(data)
        except Exception as e:
            logging.exception(e)

        try:
            if self.__tracer == 'zipkin':
                yield super(TracerTransport, self)._send()
            else:
                self._queue = []
        except Exception as e:
            logging.exception(e)

    @async_decorator
    def _send_to_statsd(self, data):
        if self.stats:
            for rec in data:
                tags = []
                t = rec['tags']
                if azc.HTTP_PATH in t and 'kind' in rec:
                    name = 'http'
                    if rec["kind"] == 'SERVER':
                        tags.append(('kind', 'in'))
                    else:
                        tags.append(('kind', 'out'))

                    copy_tags = {
                        azc.HTTP_STATUS_CODE: 'status',
                        azc.HTTP_METHOD: 'method',
                        azc.HTTP_HOST: 'host',
                        'api.key': 'api_key',
                        'api.method': 'api_method',
                        'api.code': 'api_code',
                        'api_merc.code': 'api_merc_code',
                        'peer.host': 'host',
                        'http.method': 'method',
                        'http.status': 'status',
                    }
                    for tag_key, tag_name in copy_tags.items():
                        if tag_key in t:
                            tags.append((tag_name, t[tag_key]))

                elif rec['name'].lower().startswith('db:'):
                    name = 'db'
                    tags.append(('kind', rec['name'][len('db:'):]))
                elif rec['name'].lower().startswith('redis:'):
                    name = 'redis'
                    tags.append(('kind', rec['name'][len('redis:'):]))
                elif rec['name'].lower().startswith('amqp:'):
                    name = 'amqp'
                    if rec["kind"] == 'SERVER':
                        tags.append(('kind', 'in'))
                    else:
                        tags.append(('kind', 'out'))

                elif rec['name'].lower() == 'sleep':
                    name = 'sleep'
                else:
                    name = rec['name']

                name = name.replace(' ', '_')

                if len(tags) > 0:
                    for tag in tags:
                        t = tag[1].replace(':', '-')
                        t = STATS_CLEAN_TAG_RE.sub('', t)
                        name += ',' + tag[0] + "=" + t
                self.stats.send_timer(name, int(round(rec["duration"] / 1000)))
                self.stats.send_timer('op', 1000)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Модуль доступа к вызываемой точке

"""
from collections import namedtuple
from copy import copy
from functools import partial
from inspect import signature, Parameter
from typing import (
    NamedTuple,
    Callable,
    Sequence,
    Dict,
    Any,
)

from tracuni.misc.select_coroutine import async_decorator
from tracuni.point.span_general import SpanGeneral
from tracuni.misc.helper import compose_own_call_stack
from tracuni.define.const import (
    PATH_IGNORE_METHODS,
    PATH_IGNORE_PIECE,
    COLLECT_TRACE,
)
from tracuni.define.type import (
    PointContext,
    Stage,
)


class PointAccessor:
    """Доступ к точке вызова и связанным с ней данным

    * Хранит контекст: ссылки на точку, аргументы, сокращённый стек вызовов
    и схему варианта точки
    * Преобразует все аргументы к именованным, основываясь на сигнатуре
    точки и переданным именованным аргументам для упрощения доступа к ним
    * Копирует записываемое в трассер содержимое аргументов перед их
    модификацией
    * Регистрирует сигнал о достижении определённой фазы, вызывает извлечение
    данных этой фазы по схеме варианта
    * Хранит извлеченные данные в определённой структуре, предназначенной
    для записи в трассер и для переиспользования на последующих шагах работы
    экстрактора
    * Модифицирует ограниченное множество аргументов (заголовки HTTP)
    * Вызывает точку с оригинальными и модифицированными аргументами
    """

    def __init__(self,
                 original_args: Sequence,
                 original_kwargs: Dict,
                 point_context: PointContext):
        self._positional_names = []
        self._original_args = original_args
        self._original_kwargs = original_kwargs
        self._applied_point = None
        self._done = False

        point_args = self._compose_kwargs(point_context.point_ref)

        # noinspection PyProtectedMember
        self.context = PointContext(**{
            **point_context._asdict(),
            "point_args": point_args,
        })
        self._span_general = SpanGeneral(
            self.context.engine.schema_feed.variant.span_side,
            self,
        )

    def stack_frames_to_context(self, current_frame):
        point_coro = self.context.point_ref
        stack_frames = (
            # первым элементом - название вызываемого метода
            {
                'full_path': point_coro.__qualname__,
                'short_path': point_coro.__name__,
            },
        )
        if COLLECT_TRACE:
            # собрать данные по стэку вызовов в данной точке
            # inspect.getouterframes(inspect.currentframe())

            frame_list = []
            frame = current_frame
            while frame:
                frame_list.append(frame)
                frame = frame.f_back
            stack_frames = (
                # добавить название вызываемого метода
                *stack_frames,
                *compose_own_call_stack(
                    PATH_IGNORE_METHODS,  # пропустить методы с именами
                    PATH_IGNORE_PIECE,  # пропустить модули по такому пути
                    result_length=5,  # общий размер списка не превышает
                    from_index=1,  # начинать просмотр стэка с позиции
                    look_up_stop_idx=10,  # не просматривать дальше позиции
                    frame_list=frame_list,  # стэк
                )
            )
        # noinspection PyProtectedMember
        self.context = PointContext(**{
            **self.context._asdict(),
            "call_stack": stack_frames,
        })

    def apply_point(self):
        kwargs = copy(self._original_kwargs)
        # noinspection PyProtectedMember
        kwargs.update(self.context.point_args._asdict())
        kwargs.pop('method_arguments_without_names')
        args = list(self._original_args)
        for idx, pos_name in enumerate(
            self._positional_names[:len(self._original_args)]
        ):
            args[idx] = kwargs.pop(pos_name, None)

        self._applied_point = partial(self.context.point_ref, *args, **kwargs)

    @async_decorator
    def run(self):
        # если обертка точки создана вызваем через неё
        result = None
        if self.context.span and self._applied_point:
            result = yield from self._applied_point()
            self._done = True
            self.update_context({"point_result": result})
        return result

    def has_not_been_done(self):
        return not self._done

    @async_decorator
    def register_stage(self, stage: Stage):
        yield from self._span_general.run_stage(stage)

    @async_decorator
    def register_error(self, e):
        yield from self._span_general.register_error(e)

    @async_decorator
    def register_own_error(self, e):
        self._span_general.emergency_exit(e)

    def update_context(self, new_data):
        # noinspection PyProtectedMember
        self.context = PointContext(**{
            **self.context._asdict(),
            **new_data,
        })

    def _compose_kwargs(self, point_ref: Callable) -> NamedTuple:
        sig = signature(point_ref).parameters
        self._positional_names = arg_names = [
            el
            for el in sig
            if el not in ('args', 'kwargs')
        ]
        kwargs = copy(self._original_kwargs)
        args = copy(self._original_args)
        kwargs.update(zip(arg_names, args))

        param_desc = sig
        for param in param_desc:
            pass_value = kwargs.get(param, param_desc.get(param).default)
            if pass_value != Parameter.empty:
                kwargs[param] = pass_value

        kwargs_l = len(kwargs)
        kwargs.update({'method_arguments_without_names': None})
        if kwargs_l < len(args):
            kwargs['method_arguments_without_names'] = args[kwargs_l:]

        point_args = namedtuple(
            'PointArguments',
            kwargs
        )(**kwargs)

        return point_args



import logging
import re
from urllib.parse import urlparse

# noinspection PyPackageRequirements
import statsd

from tracuni.misc.select_coroutine import (
    IOLoop,
    PeriodicCallback,
)
from tracuni.define.type import (
    StatsdOptions,
)
if not IOLoop:
    import asyncio

ERROR_NOT_CONNECTED = "StatsD is not available!"
MSG_RESPONSE = '%s.response,code=%d'
MSG_EXCEPTION = '%s.exception,code=%d'
MSG_REQUEST = '%s.http_request,url=%s,code=%d'
MSG_METHOD_APIKEY = '%s.method_apikey,method=%s,code=%d,apikey=%s'
check_http_prefix = re.compile(r'^http')


def run_later_if_enabled(fn):
    def decorated(self, *args, **kwargs):
        def wrapper():
            try:
                fn(self, *args, **kwargs)
            except Exception as e:
                logging.error(e)

        if self.enable:
            if self.run_later:
                if IOLoop:
                    IOLoop.current().spawn_callback(
                        wrapper
                    )
                else:
                    asyncio.ensure_future(wrapper())
            else:
                return wrapper()
    return decorated


class StatsD:

    def __init__(self, config):
        self.run_later = config.get('run_later', True)
        self.enable = config.get('enable', False)
        self.logging = config.get('logging', False)
        self.config = {
            'host': config.get('host', 'localhost'),
            'port': config.get('port', 8125),
            'sample_rate': 1,
            'disabled': not self.enable,
        }
        self.name = config.get('prefix', 'svc_api')
        self.hb = None
        self.hb_interval = (config.get('hb_interval', 30) or 30) * 1000

    def connect(self):
        if not self.logging:
            logger_timer = logging.getLogger(
                'statsd.client.Timer'
            )
            logger_timer.propagate = False
            logger_counter = logging.getLogger(
                'statsd.client.Counter'
            )
            logger_counter.propagate = False
            logger_connection = logging.getLogger(
                'statsd.connection.Connection'
            )
            logger_connection.propagate = False
            logger_connection.setLevel(logging.CRITICAL)

        try:
            statsd.Connection.set_defaults(**self.config)
            statsd.Client(name=self.name)
            self.hb = statsd.Counter(self.name)
            if PeriodicCallback:
                PeriodicCallback(
                    callback=self.heartbeat,
                    callback_time=self.hb_interval
                ).start()
            else:
                loop = asyncio.get_event_loop()
                task = asyncio.ensure_future(self.heartbeat())
                loop.call_later(self.hb_interval, task.cancel)
        except Exception as e:
            self.enable = False
            logging.error(e)
            logging.error(ERROR_NOT_CONNECTED)
        return

    def heartbeat(self):
        if self.enable:
            self.hb.increment('heartbeat')

    @run_later_if_enabled
    def send_gauge(self, key, value):
        gauge = statsd.Gauge(self.name)
        gauge.send(key, value)

    @run_later_if_enabled
    def exception(self, code):
        statsd.increment(MSG_EXCEPTION % (self.name, code))

    @run_later_if_enabled
    def response(self, code):
        statsd.increment(MSG_RESPONSE % (self.name, code))

    @run_later_if_enabled
    def http_request(self, url, code):
        statsd.increment(MSG_REQUEST % (self.name, url, code))

    @run_later_if_enabled
    def method_apikey(self, method, code, apikey):
        statsd.increment(MSG_METHOD_APIKEY % (self.name, method, code, apikey))

    def start_timer(self):
        if self.enable:
            timer = statsd.Timer(self.name)
            timer.start()
            return timer

    @run_later_if_enabled
    def stop_timer(self, timer, label):
        if timer:
            timer.stop(label)

    @run_later_if_enabled
    def send_timer(self, name, value):
        timer = statsd.Timer(self.name)
        timer.send(name, value)

    @staticmethod
    def read_statsd_configuration(statsd_conf):
        statsd_enable = bool(statsd_conf.enable)
        statsd_url = statsd_conf.url
        statsd_url = urlparse(
            statsd_url
            if check_http_prefix.match(statsd_url)
            else 'https://{}'.format(statsd_url)
        )

        statsd_is_configured = bool(statsd_url) and statsd_enable
        if statsd_is_configured and not statsd_url.hostname:
            statsd_is_configured = False
            logging.error('Incorrect statsd url')

        # noinspection PyProtectedMember
        updated_conf = statsd_conf._asdict()
        updated_conf.update({
            'enable': statsd_is_configured,
            'host': statsd_url.hostname,
            'port': statsd_url.port,
        })
        statsd_conf = StatsdOptions(**updated_conf)

        return statsd_conf



import json

from app.core.db import DB


class DbTasks(DB):
    async def q1(self, name, task):
        sub_tasks = task.get('tasks', {})
        sql = """
            SELECT *
            FROM
                task_manager.tasks t
            WHERE
                NOT t.closed
            LIMIT 1
        """
        res = await self.query_all(sql)
        res = [dict(el) for el in res]
        return {
            "name": name,
            "ok": True,
            "tasks": sub_tasks,
            "result": res,
        }
    async def q2(self, name, task):
        sub_tasks = task.get('tasks', {})
        sql = """
            SELECT *
            FROM
                task_manager.tasks t
            WHERE
                NOT t.closed
            LIMIT 2
        """
        res = await self.query_all(sql)
        res = [dict(el) for el in res]
        return {
            "name": name,
            "ok": True,
            "tasks": sub_tasks,
            "result": res,
        }
    async def q3(self, name, task):
        sub_tasks = task.get('tasks', {})
        sql = """
            SELECT *
            FROM
                task_manager.tasks t
            WHERE
                NOT t.closed
            LIMIT 3
        """
        res = await self.query_all(sql)
        res = [dict(el) for el in res]
        return {
            "name": name,
            "ok": True,
            "tasks": sub_tasks,
            "result": res,
        }
